#include "xparameters.h"
#include "xtmrctr.h"
#define PI 843314856       
#define AG_CONST 163008218

// Note on <32,28> fixed point precision:
//
// <32,28> is a fixed point number with the weight of the lsb equal to 2^-28
// The weight of the msb in a <32,28> is -2^4 (two's complement).
// Thus, the value '1' is represented as (1 << 28)
// Similarly, the fixed point number (1 << 26) corresponds to the value 0.25
//
// This simple scaling allows fractional precision with integer calculations

static const int angles[] = {
  210828714,
  124459457,
   65760959,
   33381289,
   16755421,
    8385878,
    4193962,
    2097109,
    1048570,
     524287,
     262143,
     131071,
      65535,
      32767,
      16383,
       8191,
       4095,
       2047,
       1024,
        511 };

#define DATALEN 65536

int target_angle[DATALEN];
int result_X[DATALEN];
int result_Y[DATALEN];

void prepare_angle() {
  unsigned rnstate = 1;                                                          
  unsigned a = 0;                                                              
  int i;                                                                       
  for (i=0; i<DATALEN; i++) {                                                
    rnstate = (rnstate >> 1) ^ (-(signed int)(rnstate & 1) & 0xd0000001u);     
    target_angle[i] = (int) rnstate % (PI/2);
  }                                                                            
}

void golden_cordic(int target, int *rX, int *rY) {
  int X, Y, T, current;
  unsigned step;
  X       = AG_CONST;  
  Y       = 0;	       
  current = 0;
  for(step=0; step < 20; step++) {
    if (target > current) {
      T          =  X - (Y >> step);
      Y          = (X >> step) + Y;
      X          = T;
      current   += angles[step];
    } else {
      T          = X + (Y >> step);
      Y          = -(X >> step) + Y;
      X          = T;
      current   -= angles[step]; 
    }
  }
  *rX = X;
  *rY = Y;
}

void your_cordic() {
  int i;
  for (i=0; i<DATALEN; i++)
    golden_cordic(target_angle[i], &(result_X[i]), &(result_Y[i]));
}

void check_result() {
  int X, Y;
  int dX, dY;
  int errors = 0;
  int i;
  int sum_abs_error = 0;
  for (i=0; i<DATALEN; i++) {                                                
    golden_cordic(target_angle[i], &X, &Y);
    dX = X - result_X[i];
    dY = Y - result_Y[i];
    if (dX) 
      errors++;
    if (dY) 
      errors++;
    sum_abs_error += (dX > 0) ? dX : -dX;
    sum_abs_error += (dY > 0) ? dY : -dY;
  }
  xil_printf("Errors: %d\n", errors);
  xil_printf("Sum_abs_error: %d\n", sum_abs_error);

}

int main(void) {
	XTmrCtr T;
	
	XTmrCtr_Initialize(&T, XPAR_OPB_TIMER_0_DEVICE_ID);
	XTmrCtr_SetResetValue(&T, 0, 0);
   prepare_angle();


   XTmrCtr_Reset(&T, 0);
   XTmrCtr_Start(&T, 0);
   //------- Measure the execution time of the function your_cordic -----

   your_cordic();

   //---------------------------------------------------------------------
   XTmrCtr_Stop(&T, 0);
   xil_printf("Cycles: %x\n", XTmrCtr_GetValue(&T, 0));

   check_result();
   return 0; 
}
